// TODO: Put public facing types in this file.

enum IterateType { year, month, day }

extension _IterateTypeExtension on IterateType {
  bool get isDay {
    return this == IterateType.day ? true : false;
  }

  bool get isMonth {
    return this == IterateType.month ? true : false;
  }

  bool get isYear {
    return this == IterateType.year ? true : false;
  }
}

enum _Order { ascending, descending }

extension _OrderExtension on _Order {
  int get add {
    return this == _Order.ascending ? 1 : -1;
  }

  bool get isAscending {
    return this == _Order.ascending ? true : false;
  }

  bool get isDescending {
    return this == _Order.descending ? true : false;
  }
}

/// DaTeTime Iterable Class
class DateTimeIterable extends Iterable<DateTime> {
  /// DateTime nesneleri üzerinde iterate edilebilen bir nesne döner.
  DateTimeIterable(DateTime start, DateTime end,
      {this.iterateType = IterateType.day, this.includeEnd = false}) {
    if (iterateType.isDay) {
      _start = DateTime(start.year, start.month, start.day);
      _end = DateTime(end.year, end.month, end.day);
    } else if (iterateType.isMonth) {
      _start = DateTime(start.year, start.month);
      _end = DateTime(end.year, end.month);
    } else {
      _start = DateTime(start.year);
      _end = DateTime(end.year);
    }
  }

  late final DateTime _start;
  late final DateTime _end;
  final bool includeEnd;
  final IterateType iterateType;

  @override
  Iterator<DateTime> get iterator =>
      _DateTimeIterator(_start, _end, iterateType, includeEnd);
}

class _DateTimeIterator implements Iterator<DateTime> {
  _DateTimeIterator(this.start, this.end, this.iterateType, this.includeEnd) {
    _indexDate = start;
    _currentDateTime = start;
    _order = _findOrder();
  }

  final DateTime start;
  final DateTime end;
  final IterateType iterateType;
  final bool includeEnd;
  late final _Order _order;

  late DateTime _currentDateTime;
  late DateTime _indexDate;

  _Order _findOrder() {
    return start.isBefore(end) ? _Order.ascending : _Order.descending;
  }

  @override
  DateTime get current => _currentDateTime;

  @override
  bool moveNext() {
    // if last value not included then return
    if ((_indexDate == end && !includeEnd) ||
        (_indexDate.isAfter(end) && _order.isAscending) ||
        (_indexDate.isBefore(end) && _order.isDescending)) {
      return false;
    }

    switch (iterateType) {
      case IterateType.day:
        moveNextDay();
        return true;
      case IterateType.month:
        moveNextMonth();
        return true;
      case IterateType.year:
        moveNextYear();
        return true;
    }
  }

  bool moveNextDay() {
    _currentDateTime = _indexDate;
    _indexDate = _indexDate.add(Duration(days: _order.add));
    return true;
  }

  bool moveNextMonth() {
    _currentDateTime = _indexDate;
    _indexDate = DateTime(_indexDate.year, (_indexDate.month + _order.add));
    return true;
  }

  bool moveNextYear() {
    _currentDateTime = _indexDate;
    _indexDate = DateTime(_indexDate.year + _order.add);
    return true;
  }
}
